import logging
import os
import sys
import click
from functools import partial

from .config import load_config
from .models import Snapshot, Peer, Table, Base
from .operations import (
    copy_database,
    create_database,
    database_exists,
    remove_database,
    rename_database,
    terminate_database_connections,
    list_of_databases,
    import_database
)
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.exc import ProgrammingError
from psutil import pid_exists


__version__ = '0.4.4a'
logger = logging.getLogger(__name__)


class Operations(object):
    def __init__(self, raw_connection, config):
        self.terminate_database_connections = partial(
            terminate_database_connections, raw_connection
        )
        self.create_database = partial(create_database, raw_connection)
        self.copy_database = partial(copy_database, raw_connection)
        self.database_exists = partial(database_exists, raw_connection)
        self.rename_database = partial(rename_database, raw_connection)
        self.remove_database = partial(remove_database, raw_connection)
        self.list_of_databases = partial(list_of_databases, raw_connection)
        self.import_database = partial(import_database, raw_connection)

class Stellar(object):
    def __init__(self):
        logger.debug('Initialized Stellar()')
        self.load_config()
        self.init_database()

    def load_config(self):
        self.config = load_config()
        logging.basicConfig(level=self.config['logging'])

    def init_database(self):
        self.raw_db = create_engine(self.config['url'], echo=False)
        self.raw_conn = self.raw_db.connect()
        self.operations = Operations(self.raw_conn, self.config)

        try:
            self.raw_conn.connection.set_isolation_level(0)
        except AttributeError:
            logger.info('Could not set isolation level to 0')

        self.db = create_engine(self.config['stellar_url'], echo=False)
        self.db.session = sessionmaker(bind=self.db)()
        self.raw_db.session = sessionmaker(bind=self.raw_db)()
        tables_missing = self.create_stellar_database()

        self.create_stellar_tables()

        # logger.getLogger('sqlalchemy.engine').setLevel(logger.WARN)


    def init_remote_database(self, peer_url):
        self.remote_raw_db = create_engine(peer_url, echo=False)
        self.remote_raw_conn = self.remote_raw_db.connect()
        self.remote_operations = Operations(self.remote_raw_conn, self.config)

        try:
            self.remote_raw_conn.connection.set_isolation_level(0)
        except AttributeError:
            logger.info('Could not set isolation level to 0 on remote peer')

        self.remote_db = create_engine(peer_url, echo=False)
        self.remote_db.session = sessionmaker(bind=self.remote_db)()
        self.remote_raw_db.session = sessionmaker(bind=self.remote_raw_db)()
        tables_missing = self.create_stellar_database()

        self.create_stellar_tables()


    def create_stellar_database(self):
        if not self.operations.database_exists('stellar_data'):
            self.operations.create_database('stellar_data')
            return True
        else:
            return False

    def create_stellar_tables(self):
        Base.metadata.create_all(self.db)
        self.db.session.commit()

    def get_peer(self, peer_name):
        try:
            peer = self.db.session.query(Peer).filter(
                Peer.peer_name == peer_name,
                ).first()
        except:
            click.echo('Fetching peer "%s" failed.' % name)
            exit(1)
        return peer


    def get_peers(self):
        return self.db.session.query(Peer).all()

    def create_peer(self, peer_name, url, regex):
        if self.get_peer(peer_name):
            click.echo("Peer (%s) already exists." % peer_name)
            sys.exit(1)
        peer = Peer(
            peer_name=peer_name,
            url=url,
            regex=regex
        )
        self.db.session.add(peer)
        self.db.session.flush()
        self.db.session.commit()

    def remove_peer(self, peer):
        if not self.get_peer(peer_name):
            click.echo("Peer (%s) does not exist." % peer_name)
            sys.exit(1)
        self.db.session.delete(peer)
        self.db.session.commit()

    def get_remote_snapshots(self, peer_name):
        if not self.get_peer(peer_name):
            click.echo("Peer (%s) does not exist." % peer_name)
            sys.exit(1)

        peer_url = self.db.session.query(Peer).filter(Peer.peer_name == peer_name,).first().url
        self.init_remote_database(peer_url)

        return self.remote_db.session.query(Snapshot).filter(
            Snapshot.project_name == self.config['project_name']
        ).order_by(
            Snapshot.created_at.desc()
        ).all()

    def get_remote_snapshot(self, peer_name, snapshot_name):
        if not self.get_peer(peer_name):
            click.echo("Peer (%s) does not exist." % peer_name)
            sys.exit(1)

        peer_url = self.db.session.query(Peer).filter(Peer.peer_name == peer_name,).first().url
        self.init_remote_database(peer_url)

        return self.remote_db.session.query(Snapshot).filter(
            Snapshot.snapshot_name == snapshot_name,
            Snapshot.project_name == self.config['project_name']
        ).first()

    def import_remote_snapshot(self, peer_name, remote_snapshot, before_copy=None):
        if not self.get_peer(peer_name):
            click.echo("Peer (%s) does not exist." % peer_name)
            sys.exit(1)

        if not remote_snapshot:
            click.echo("An error occured! Snapshot %s is not valid." % remote_snapshot)
            sys.exit(1)

        remote_url = self.db.session.query(Peer).filter(Peer.peer_name == peer_name,).first().url
        sed_regex = self.db.session.query(Peer).filter(Peer.peer_name == peer_name,).first().regex
        local_url = self.config['url']
        snapshot = Snapshot(
            snapshot_name = remote_snapshot.snapshot_name,
            project_name = remote_snapshot.project_name,
            hash = remote_snapshot.hash,
            created_at = remote_snapshot.created_at
        )

        click.echo("Importing snapshot %s" % remote_snapshot.snapshot_name)
        for table in remote_snapshot.tables:
            if not self.remote_operations.database_exists(
                table.get_table_name('slave')
            ):
                click.echo(
                    "Database %s does not exist."
                    % table.get_table_name('slave')
                )
                sys.exit(1)
            try:
                self.operations.import_database(
                    remote_url,
                    table.get_table_name('slave'),
                    local_url,
                    table.get_table_name('slave'),
                    sed_regex
                    )
            except ProgrammingError:
                logger.warn('Database %s does not exist.' % table.get_table_name('slave'))
           
            if not self.operations.database_exists(
                table.get_table_name('slave')
            ):
                click.echo(
                    "Database import failed for %s."
                    % table.get_table_name('slave')
                )
                sys.exit(1)

            if self.operations.database_exists(
                table.get_table_name('master')
            ):
                self.operations.remove_database(
                    table.get_table_name('master')
                )
            self.operations.rename_database(
                table.get_table_name('slave'),
                table.get_table_name('master')
            )     

        self.db.session.add(snapshot)
        self.db.session.commit()
        self.inline_slave_copy(snapshot)
        click.echo("Done.")

    def get_snapshot(self, snapshot_name):
        return self.db.session.query(Snapshot).filter(
            Snapshot.snapshot_name == snapshot_name,
            Snapshot.project_name == self.config['project_name']
        ).first()

    def get_snapshots(self):
        return self.db.session.query(Snapshot).filter(
            Snapshot.project_name == self.config['project_name']
        ).order_by(
            Snapshot.created_at.desc()
        ).all()

    def get_latest_snapshot(self):
        return self.db.session.query(Snapshot).filter(
            Snapshot.project_name == self.config['project_name']
        ).order_by(Snapshot.created_at.desc()).first()

    def create_snapshot(self, snapshot_name, before_copy=None):
        snapshot = Snapshot(
            snapshot_name=snapshot_name,
            project_name=self.config['project_name']
        )
        self.db.session.add(snapshot)
        self.db.session.flush()

        for table_name in self.config['tracked_databases']:
            if before_copy:
                before_copy(table_name)
            table = Table(
                table_name=table_name,
                snapshot=snapshot
            )
            logger.debug('Copying %s to %s' % (
                table_name,
                table.get_table_name('master')
            ))
            self.operations.copy_database(
                table_name,
                table.get_table_name('master')
            )
            self.db.session.add(table)
        self.db.session.commit()

        self.start_background_slave_copy(snapshot)

    def remove_snapshot(self, snapshot):
        for table in snapshot.tables:
            try:
                self.operations.remove_database(
                    table.get_table_name('master')
                )
            except ProgrammingError:
                pass
            try:
                self.operations.remove_database(
                    table.get_table_name('slave')
                )
            except ProgrammingError:
                pass
            self.db.session.delete(table)
        self.db.session.delete(snapshot)
        self.db.session.commit()

    def rename_snapshot(self, snapshot, new_name):
        snapshot.snapshot_name = new_name
        self.db.session.commit()

    def restore(self, snapshot):
        for table in snapshot.tables:
            click.echo("Restoring database %s" % table.table_name)
            if not self.operations.database_exists(
                table.get_table_name('slave')
            ):
                click.echo(
                    "Database %s does not exist."
                    % table.get_table_name('slave')
                )
                sys.exit(1)
            try:
                self.operations.remove_database(table.table_name)
            except ProgrammingError:
                logger.warn('Database %s does not exist.' % table.table_name)
            self.operations.rename_database(
                table.get_table_name('slave'),
                table.table_name
            )
        snapshot.worker_pid = 1
        self.db.session.commit()

        self.start_background_slave_copy(snapshot)

    def start_background_slave_copy(self, snapshot):
        logger.debug('Starting background slave copy')
        snapshot_id = snapshot.id

        self.raw_conn.close()
        self.raw_db.session.close()
        self.db.session.close()

        pid = os.fork() if hasattr(os, 'fork') else None
        if pid:
            return

        self.init_database()
        self.operations = Operations(self.raw_conn, self.config)

        snapshot = self.db.session.query(Snapshot).get(snapshot_id)
        snapshot.worker_pid = os.getpid()
        self.db.session.commit()
        self.inline_slave_copy(snapshot)
        sys.exit()

    def inline_slave_copy(self, snapshot):
        for table in snapshot.tables:
            self.operations.copy_database(
                table.get_table_name('master'),
                table.get_table_name('slave')
            )
        snapshot.worker_pid = None
        self.db.session.commit()

    def is_copy_process_running(self, snapshot):
        return pid_exists(snapshot.worker_pid)

    def is_old_database(self):
        for snapshot in self.db.session.query(Snapshot):
            for table in snapshot.tables:
                for postfix in ('master', 'slave'):
                    old_name = table.get_table_name(postfix=postfix, old=True)
                    if self.operations.database_exists(old_name):
                        return True
        return False

    def update_database_names_to_new_version(self, after_rename=None):
        for snapshot in self.db.session.query(Snapshot):
            for table in snapshot.tables:
                for postfix in ('master', 'slave'):
                    old_name = table.get_table_name(postfix=postfix, old=True)
                    new_name = table.get_table_name(postfix=postfix, old=False)
                    if self.operations.database_exists(old_name):
                        self.operations.rename_database(old_name, new_name)
                        if after_rename:
                            after_rename(old_name, new_name)

    def delete_orphan_snapshots(self, after_delete=None):
        stellar_databases = set()
        for snapshot in self.db.session.query(Snapshot):
            for table in snapshot.tables:
                stellar_databases.add(table.get_table_name('master'))
                stellar_databases.add(table.get_table_name('slave'))

        databases = set(self.operations.list_of_databases())

        for database in filter(
            lambda database: (
                database.startswith('stellar_') and
                database != 'stellar_data'
            ),
            (databases-stellar_databases)
        ):
            self.operations.remove_database(database)
            if after_delete:
                after_delete(database)

    @property
    def default_snapshot_name(self):
        n = 1
        while self.db.session.query(Snapshot).filter(
            Snapshot.snapshot_name == 'snap%d' % n,
            Snapshot.project_name == self.config['project_name']
        ).count():
            n += 1
        return 'snap%d' % n



import click
import logging
import sqlalchemy_utils
from .cmd import (
    get_temp,
    remove_temp,
    dump_database,
    restore_database,
    apply_regex)
from urllib.parse import urlparse, parse_qs


logger = logging.getLogger(__name__)


SUPPORTED_DIALECTS = (
    'postgresql',
    'mysql'
)

class NotSupportedDatabase(Exception):
    pass


def get_engine_url(raw_conn, database):
    url = str(raw_conn.engine.url)
    if url.count('/') == 3 and url.endswith('/'):
        return '%s%s' % (url, database)
    else:
        if not url.endswith('/'):
            url += '/'
        return '%s/%s' % ('/'.join(url.split('/')[0:-2]), database)


def _get_pid_column(raw_conn):
    version = [int(x) for x in raw_conn.execute('SHOW server_version;').first()[0].split('.')]
    return 'pid' if version >= [9, 2] else 'procpid'


def terminate_database_connections(raw_conn, database):
    logger.debug('terminate_database_connections(%r)', database)
    if raw_conn.engine.dialect.name == 'postgresql':
        pid_column = _get_pid_column(raw_conn)

        raw_conn.execute(
            '''
                SELECT pg_terminate_backend(pg_stat_activity.%(pid_column)s)
                FROM pg_stat_activity
                WHERE
                    pg_stat_activity.datname = '%(database)s' AND
                    %(pid_column)s <> pg_backend_pid();
            ''' % {'pid_column': pid_column, 'database': database}
        )
    else:
        # NotYetImplemented
        pass


def create_database(raw_conn, database):
    logger.debug('create_database(%r)', database)
    return sqlalchemy_utils.functions.create_database(
        get_engine_url(raw_conn, database)
    )


def copy_database(raw_conn, from_database, to_database):
    logger.debug('copy_database(%r, %r)', from_database, to_database)
    terminate_database_connections(raw_conn, from_database)

    if raw_conn.engine.dialect.name == 'postgresql':
        raw_conn.execute(
            '''
                CREATE DATABASE "%s" WITH TEMPLATE "%s";
            ''' %
            (
                to_database,
                from_database
            )
        )
    elif raw_conn.engine.dialect.name == 'mysql':
        # Horribly slow implementation.
        create_database(raw_conn, to_database)
        for row in raw_conn.execute('SHOW TABLES in %s;' % from_database):
            raw_conn.execute('''
                CREATE TABLE %s.%s LIKE %s.%s
            ''' % (
                to_database,
                row[0],
                from_database,
                row[0]
            ))
            raw_conn.execute('ALTER TABLE %s.%s DISABLE KEYS' % (
                to_database,
                row[0]
            ))
            raw_conn.execute('''
                INSERT INTO %s.%s SELECT * FROM %s.%s
            ''' % (
                to_database,
                row[0],
                from_database,
                row[0]
            ))
            raw_conn.execute('ALTER TABLE %s.%s ENABLE KEYS' % (
                to_database,
                row[0]
            ))
    else:
        raise NotSupportedDatabase()


def database_exists(raw_conn, database):
    logger.debug('database_exists(%r)', database)
    return sqlalchemy_utils.functions.database_exists(
        get_engine_url(raw_conn, database)
    )


def remove_database(raw_conn, database):
    logger.debug('remove_database(%r)', database)
    terminate_database_connections(raw_conn, database)
    return sqlalchemy_utils.functions.drop_database(
        get_engine_url(raw_conn, database)
    )


def rename_database(raw_conn, from_database, to_database):
    logger.debug('rename_database(%r, %r)', from_database, to_database)
    terminate_database_connections(raw_conn, from_database)
    if raw_conn.engine.dialect.name == 'postgresql':
        raw_conn.execute(
            '''
                ALTER DATABASE "%s" RENAME TO "%s"
            ''' %
            (
                from_database,
                to_database
            )
        )
    elif raw_conn.engine.dialect.name == 'mysql':
        create_database(raw_conn, to_database)
        for row in raw_conn.execute('SHOW TABLES in %s;' % from_database):
            raw_conn.execute('''
                RENAME TABLE %s.%s TO %s.%s;
            ''' % (
                from_database,
                row[0],
                to_database,
                row[0]
            ))
        remove_database(raw_conn, from_database)
    else:
        raise NotSupportedDatabase()


def list_of_databases(raw_conn):
    logger.debug('list_of_databases()')
    if raw_conn.engine.dialect.name == 'postgresql':
        return [
            row[0]
            for row in raw_conn.execute('''
                SELECT datname FROM pg_database
                WHERE datistemplate = false
            ''')
        ]
    elif raw_conn.engine.dialect.name == 'mysql':
        return [
            row[0]
            for row in raw_conn.execute('''SHOW DATABASES''')
        ]
    else:
        raise NotSupportedDatabase()


def import_database(raw_conn, remote_url, from_database, local_url, to_database, regex=None):
    remote = {
        'username': urlparse(remote_url).username,
        'password': urlparse(remote_url).password,
        'hostname': urlparse(remote_url).hostname,
        'port': urlparse(remote_url).port
        }
    if not remote['username']:
            click.echo("Username missing from peer URL.")
            sys.exit(1)
    if not remote['password']:
            click.echo("Password missing from peer URL.")
            sys.exit(1)

    local = {
        'username': urlparse(local_url).username,
        'password': urlparse(local_url).password,
        'hostname': urlparse(local_url).hostname,
        'port': urlparse(local_url).port,
        }
    if not local['username']:
            click.echo("Username missing from local URL.")
            sys.exit(1)
    if not local['password']:
            click.echo("Password missing from local URL.")
            sys.exit(1)

    if raw_conn.engine.dialect.name == 'postgresql':
        tmp = get_temp()
        click.echo("Dumping from database %s to file %s." % (from_database, tmp))
        stdout,stderr = dump_database(
            remote['hostname'],
            remote['port'],
            from_database,
            remote['username'],
            remote['password'],
            tmp)
        # if stdout:
        #     click.echo('%s' % stdout)
        if stderr:
            click.echo('%s' % stderr)

        if regex:
            click.echo("Running regex on file %s." % tmp)
            stdout, stderr = apply_regex(tmp, regex)
            if stderr:
                click.echo('%s' % stderr)
            if stdout:
                click.echo('%s' % stdout)

        click.echo("Restoring to database %s from file %s." % (to_database, tmp))    
        sqlalchemy_utils.functions.create_database(
            get_engine_url(raw_conn, to_database))    
        stdout,stderr = restore_database(
            local['hostname'],
            local['port'],
            to_database,
            local['username'],
            local['password'],
            tmp)
        # if stdout:
        #     click.echo('%s' % stdout)
        if stderr:
            click.echo('%s' % stderr)
        remove_temp(tmp)
        click.echo("Temp file %s removed." % tmp)
    else:
        raise NotSupportedDatabase()                



import pytest

from stellar.operations import _get_pid_column


class ConnectionMock(object):
    def __init__(self, version):
        self.version = version

    def execute(self, query):
        return self

    def first(self):
        return [self.version]


class TestGetPidColumn(object):
    @pytest.mark.parametrize('version', ['9.1', '8.9', '9.1.9', '8.9.9'])
    def test_returns_procpid_for_version_older_than_9_2(self, version):
        raw_conn = ConnectionMock(version=version)
        assert _get_pid_column(raw_conn) == 'procpid'

    @pytest.mark.parametrize(
        'version',
        ['9.2', '9.3', '10.0', '9.2.1', '10.1.1']
    )
    def test_returns_pid_for_version_equal_or_newer_than_9_2(self, version):
        raw_conn = ConnectionMock(version=version)
        assert _get_pid_column(raw_conn) == 'pid'



from subprocess import PIPE,Popen
import os
import shlex
import tempfile
import click


def dump_database(hostname, port, database, username, password, dstfile):
    cmd = 'pg_dump -h %s -p %s -U %s -f %s -d %s'  % (hostname, port, username, dstfile, database)
    cmd = shlex.split(cmd)
    local_env = os.environ.copy()
    local_env["PGPASSWORD"] = password
    p = Popen(cmd, shell=False, env=local_env, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    return p.communicate()

def restore_database(hostname, port, database, username, password, srcfile):
    cmd = 'psql -h %s -p %s -U %s -d %s -f %s' % (hostname, port, username, database, srcfile)
    cmd = shlex.split(cmd)
    local_env = os.environ.copy()
    local_env["PGPASSWORD"] = password
    p = Popen(cmd, shell=False,  env=local_env, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    return p.communicate()

def get_temp():
    f = tempfile.NamedTemporaryFile(delete=False)
    name = f.name
    f.close
    return name

def remove_temp(f_name):
    return os.unlink(f_name)

def apply_regex(f, regex):
    cmd = 'sed -i ''%s'' %s' % (regex, f)    
    cmd = shlex.split(cmd)
    p = Popen(cmd, shell=False, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    return p.communicate()


import os
import logging
import yaml
from schema import Use, Schema, SchemaError, Optional
from os.path import expanduser

class InvalidConfig(Exception):
    pass


class MissingConfig(Exception):
    pass


default_config = {
    'logging': 30,
    'migrate_from_0_3_2': True
}
schema = Schema({
    'stellar_url': Use(str),
    'url': Use(str),
    'project_name': Use(str),
    'tracked_databases': [Use(str)],
    Optional('logging'): int,
    Optional('migrate_from_0_3_2'): bool
})


def get_config_path():
    home_directory = expanduser("~/")
    while True:
        try:
            with open(
                os.path.join(home_directory, '.stellar.yaml'),
                'rb'
            ) as fp:
                return os.path.join(home_directory, '.stellar.yaml')
        except IOError:
            pass

        home_directory = os.path.abspath(
            os.path.join(home_directory, '..')
        )
        if home_directory == '/':
            return None


def load_config():
    config = {}
    home_directory = expanduser("~/")
    while True:
        try:
            with open(
                os.path.join(home_directory, '.stellar.yaml'),
                'rb'
            ) as fp:
                config = yaml.safe_load(fp)
                break
        except IOError:
            pass
        home_directory = os.path.abspath(
            os.path.join(home_directory, '..')
        )

        if home_directory == '/':
            break

    if not config:
        raise MissingConfig()

    for k, v in default_config.items():
        if k not in config:
            config[k] = v

    try:
        return schema.validate(config)
    except SchemaError as e:
        raise InvalidConfig(e)


def save_config(config):
    logging.getLogger(__name__).debug('save_config()')
    with open(get_config_path(), "w") as fp:
        yaml.dump(config, fp)



import pytest
import stellar
import tempfile


class TestCase(object):
    @pytest.yield_fixture(autouse=True)
    def config(self, monkeypatch):
        with tempfile.NamedTemporaryFile() as tmp:
            def load_test_config(self):
                self.config = {
                    'stellar_url': 'sqlite:///%s' % tmp.name,
                    'url': 'sqlite:///%s' % tmp.name,
                    'project_name': 'test_project',
                    'tracked_databases': ['test_database'],
                    'TEST': True
                }
                return None
            monkeypatch.setattr(stellar.app.Stellar, 'load_config', load_test_config)
            yield


class Test(TestCase):
    def test_setup_method_works(self, monkeypatch):
        monkeypatch.setattr(
            stellar.app.Stellar,
            'create_stellar_database',
            lambda x: None
        )
        app = stellar.app.Stellar()
        for key in (
            'TEST',
            'stellar_url',
            'url',
            'project_name',
            'tracked_databases',
        ):
            assert app.config[key]

    def test_shows_not_enough_arguments(self):
        with pytest.raises(SystemExit) as e:
            stellar.command.main()

    def test_app_context(self, monkeypatch):
        monkeypatch.setattr(
            stellar.app.Stellar,
            'create_stellar_database',
            lambda x: None
        )
        app = stellar.app.Stellar()



import sys
from datetime import datetime
from time import sleep

import humanize
import click
import logging
from sqlalchemy import create_engine
from sqlalchemy.exc import OperationalError

from .app import Stellar, __version__
from .config import InvalidConfig, MissingConfig, load_config, save_config
from .operations import database_exists, list_of_databases, SUPPORTED_DIALECTS


def upgrade_from_old_version(app):
    if app.config['migrate_from_0_3_2']:
        if app.is_old_database():
            click.echo('Upgrading from old Stellar version...')
            def after_rename(old_name, new_name):
                click.echo('* Renamed %s to %s' % (old_name, new_name))
            app.update_database_names_to_new_version(after_rename=after_rename)

        app.config['migrate_from_0_3_2'] = False
        save_config(app.config)

def get_app():
    app = Stellar()
    upgrade_from_old_version(app)
    return app


@click.group()
def stellar():
    """Fast database snapshots for development. It's like Git for databases."""
    pass


@stellar.command()
def version():
    """Shows version number"""
    click.echo("Stellar %s" % __version__)


@stellar.command()
def gc():
    """Deletes old stellar tables that are not used anymore"""
    def after_delete(database):
        click.echo("Deleted table %s" % database)

    app = get_app()
    upgrade_from_old_version(app)
    app.delete_orphan_snapshots(after_delete)


@stellar.command()
@click.argument('name', required=False)
def snapshot(name):
    """Takes a snapshot of the database"""
    app = get_app()
    upgrade_from_old_version(app)
    name = name or app.default_snapshot_name

    if app.get_snapshot(name):
        click.echo("Snapshot with name %s already exists" % name)
        sys.exit(1)
    else:
        def before_copy(table_name):
            click.echo("Snapshotting database %s" % table_name)
        app.create_snapshot(name, before_copy=before_copy)


@stellar.command()
def list():
    """Returns a list of snapshots"""
    snapshots = get_app().get_snapshots()

    click.echo('\n'.join(
        '%s: %s' % (
            s.snapshot_name,
            humanize.naturaltime(datetime.utcnow() - s.created_at)
        )
        for s in snapshots
    ))

@stellar.command()
def peers():
    """Returns a list of peers"""
    peers = get_app().get_peers()

    click.echo('\n'.join(
        '%s' % (p.peer_name)
        for p in peers
    ))

@stellar.command()
@click.argument('name', required=True)
def peer(name):
    """Returns a single of peer"""
    peer = get_app().get_peer(name)
    if peer:
        click.echo('%s: %s (%s)' % (peer.peer_name, peer.url, peer.regex))
    else:
        click.echo('Peer (%s) not found.' % name)

@stellar.command()
@click.argument('name', required=True)
@click.argument('url', required=True)
@click.argument('regex', required=False)
def peer_create(name, url, regex=''):
    """Creates a new peer"""
    peer = get_app().create_peer(name, url, regex)

    click.echo('Peer "%s": "%s" "%s" created.' % (name, url, regex))

@stellar.command()
@click.argument('name', required=True)
def peer_remove(name):
    """Removes a peer"""
    app = get_app()

    peer = app.get_peer(name)
    if not peer:
        click.echo("Couldn't find snapshot %s" % name)
        sys.exit(1)

    app.remove_peer(peer)
    click.echo('Peer (%s) removed.' % (name))


@stellar.command()
@click.argument('peer', required=True)
def remote_list(peer):
    """Returns a list of snapshots from a remote peer"""
    snapshots = get_app().get_remote_snapshots(peer)

    click.echo('\n'.join(
        '%s (peer: %s): %s [db: %s / %s]' % (
            s.snapshot_name,
            peer,
            humanize.naturaltime(datetime.utcnow() - s.created_at),
            ' '.join(t.get_table_name('master') for t in s.tables),
            ' '.join(t.get_table_name('slave') for t in s.tables),

        )
        for s in snapshots
    ))

@stellar.command()
@click.argument('peer', required=True)
@click.argument('snapshot_name', required=True)
def remote_get(peer, snapshot):
    """Returns a single snapshot from a remote peer"""
    snapshot = get_app().get_remote_snapshot(peer, snapshot)

    click.echo(
        '%s(%s): %s' % (
            snapshot.snapshot_name,
            peer,
            humanize.naturaltime(datetime.utcnow() - snapshot.created_at)
        )
    )
    click.echo('%s' % snapshot.hash)

@stellar.command()
@click.argument('peer', required=True)
@click.argument('snapshot', required=True)
def remote_import(peer, snapshot):
    """Imports a single snapshot from a remote peer"""
    app = get_app()
    local_snapshot = app.get_snapshot(snapshot)
    if local_snapshot:
        click.echo("Snapshot name exists locally.")
        sys.exit(1)
    remote_snapshot = app.get_remote_snapshot(peer, snapshot)
    result = app.import_remote_snapshot(peer, remote_snapshot)

@stellar.command()
@click.argument('name', required=False)
def restore(name):
    """Restores the database from a snapshot"""
    app = get_app()

    if not name:
        snapshot = app.get_latest_snapshot()
        if not snapshot:
            click.echo(
                "Couldn't find any snapshots for project %s" %
                load_config()['project_name']
            )
            sys.exit(1)
    else:
        snapshot = app.get_snapshot(name)
        if not snapshot:
            click.echo(
                "Couldn't find snapshot with name %s.\n"
                "You can list snapshots with 'stellar list'" % name
            )
            sys.exit(1)

    # Check if slaves are ready
    if not snapshot.slaves_ready:
        if app.is_copy_process_running(snapshot):
            sys.stdout.write(
                'Waiting for background process(%s) to finish' %
                snapshot.worker_pid
            )
            sys.stdout.flush()
            while not snapshot.slaves_ready:
                sys.stdout.write('.')
                sys.stdout.flush()
                sleep(1)
                app.db.session.refresh(snapshot)
            click.echo('')
        else:
            click.echo('Background process missing, doing slow restore.')
            app.inline_slave_copy(snapshot)

    app.restore(snapshot)
    click.echo('Restore complete.')


@stellar.command()
@click.argument('name')
def remove(name):
    """Removes a snapshot"""
    app = get_app()

    snapshot = app.get_snapshot(name)
    if not snapshot:
        click.echo("Couldn't find snapshot %s" % name)
        sys.exit(1)

    click.echo("Deleting snapshot %s" % name)
    app.remove_snapshot(snapshot)
    click.echo("Deleted")


@stellar.command()
@click.argument('old_name')
@click.argument('new_name')
def rename(old_name, new_name):
    """Renames a snapshot"""
    app = get_app()

    snapshot = app.get_snapshot(old_name)
    if not snapshot:
        click.echo("Couldn't find snapshot %s" % old_name)
        sys.exit(1)

    new_snapshot = app.get_snapshot(new_name)
    if new_snapshot:
        click.echo("Snapshot with name %s already exists" % new_name)
        sys.exit(1)

    app.rename_snapshot(snapshot, new_name)
    click.echo("Renamed snapshot %s to %s" % (old_name, new_name))


@stellar.command()
@click.argument('name')
def replace(name):
    """Replaces a snapshot"""
    app = get_app()

    snapshot = app.get_snapshot(name)
    if not snapshot:
        click.echo("Couldn't find snapshot %s" % name)
        sys.exit(1)

    app.remove_snapshot(snapshot)
    app.create_snapshot(name)
    click.echo("Replaced snapshot %s" % name)


@stellar.command()
def init():
    """Initializes Stellar configuration."""
    while True:
        url = click.prompt(
            "Please enter the url for your database.\n\n"
            "For example:\n"
            "PostgreSQL: postgresql://localhost:5432/\n"
            "MySQL: mysql+pymysql://root@localhost/"
        )
        if url.count('/') == 2 and not url.endswith('/'):
            url = url + '/'

        if (
            url.count('/') == 3 and
            url.endswith('/') and
            url.startswith('postgresql://')
        ):
            connection_url = url + 'template1'
        else:
            connection_url = url

        engine = create_engine(connection_url, echo=False)
        try:
            conn = engine.connect()
        except OperationalError as err:
            click.echo("Could not connect to database: %s" % url)
            click.echo("Error message: %s" % err.message)
            click.echo('')
        else:
            break

    if engine.dialect.name not in SUPPORTED_DIALECTS:
        click.echo("Your engine dialect %s is not supported." % (
            engine.dialect.name
        ))
        click.echo("Supported dialects: %s" % (
            ', '.join(SUPPORTED_DIALECTS)
        ))

    if url.count('/') == 3 and url.endswith('/'):
        while True:
            click.echo("You have the following databases: %s" % ', '.join([
                db for db in list_of_databases(conn)
                if not db.startswith('stellar_')
            ]))

            db_name = click.prompt(
                "Please enter the name of the database (eg. projectdb)"
            )
            if database_exists(conn, db_name):
                break
            else:
                click.echo("Could not find database %s" % db_name)
                click.echo('')
    else:
        db_name = url.rsplit('/', 1)[-1]
        url = url.rsplit('/', 1)[0] + '/'

    name = click.prompt(
        'Please enter your project name (used internally, eg. %s)' % db_name,
        default=db_name
    )

    raw_url = url

    if engine.dialect.name == 'postgresql':
        raw_url = raw_url + 'template1'

    with open('~/.stellar.yaml', 'w') as project_file:
        project_file.write(
            """
project_name: '%(name)s'
tracked_databases: ['%(db_name)s']
url: '%(raw_url)s'
stellar_url: '%(url)sstellar_data'
            """.strip() %
            {
                'name': name,
                'raw_url': raw_url,
                'url': url,
                'db_name': db_name
            }
        )

    click.echo("Wrote ~/.stellar.yaml")
    click.echo('')
    if engine.dialect.name == 'mysql':
        click.echo("Warning: MySQL support is still in beta.")
    click.echo("Tip: You probably want to take a snapshot: stellar snapshot")


def main():
    try:
        stellar()
    except MissingConfig:
        click.echo("You don't have ~/.stellar.yaml configuration yet.")
        click.echo("Initialize it by running: stellar init")
        sys.exit(1)
    except InvalidConfig as e:
        click.echo("Your ~/.stellar.yaml configuration is wrong: %s" % e.message)
        sys.exit(1)
    except ImportError as e:
        libraries = {
            'psycopg2': 'PostreSQL',
            'pymysql': 'MySQL',
        }
        for library, name in libraries.items():
            if 'No module named' in str(e) and library in str(e):
                click.echo(
                    "Python library %s is required for %s support." %
                    (library, name)
                )
                click.echo("You can install it with pip:")
                click.echo("pip install %s" % library)
                sys.exit(1)
            elif 'No module named' in str(e) and 'MySQLdb' in str(e):
                click.echo(
                    "MySQLdb binary drivers are required for MySQL support. "
                    "You can try installing it with these instructions: "
                    "http://stackoverflow.com/questions/454854/no-module-named"
                    "-mysqldb"
                )
                click.echo('')
                click.echo("Alternatively you can use pymysql instead:")
                click.echo("1. Install it first: pip install pymysql")
                click.echo(
                    "2. Specify database url as "
                    "mysql+pymysql://root@localhost/ and not as "
                    "mysql://root@localhost/"
                )
                sys.exit(1)
        raise

if __name__ == '__main__':
    main()



import hashlib
import uuid
from datetime import datetime

import sqlalchemy as sa
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


def get_unique_hash():
    return hashlib.md5(str(uuid.uuid4()).encode('utf-8')).hexdigest()


class Snapshot(Base):
    __tablename__ = 'snapshot'
    id = sa.Column(
        sa.Integer,
        sa.Sequence('snapshot_id_seq'),
        primary_key=True
    )
    snapshot_name = sa.Column(sa.String(255), nullable=False)
    project_name = sa.Column(sa.String(255), nullable=False)
    hash = sa.Column(sa.String(32), nullable=False, default=get_unique_hash)
    created_at = sa.Column(sa.DateTime, default=datetime.utcnow)
    worker_pid = sa.Column(sa.Integer, nullable=True)

    @property
    def slaves_ready(self):
        return self.worker_pid is None

    def __repr__(self):
        return "<Snapshot(snapshot_name=%r)>" % (
            self.snapshot_name
        )

class Peer(Base):
    __tablename__ = 'peer'
    id = sa.Column(
        sa.Integer,
        sa.Sequence('peer_id_seq'),
        primary_key=True
    )
    peer_name = sa.Column(sa.String(255), nullable=False)
    url = sa.Column(sa.String(255), nullable=False)
    regex = sa.Column(sa.String(255), nullable=True)

    @property
    def slaves_ready(self):
        return self.worker_pid is None

    def __repr__(self):
        return "<Peer(peer_name=%r)>" % (
            self.peer_name
        )

class Table(Base):
    __tablename__ = 'table'
    id = sa.Column(sa.Integer, sa.Sequence('table_id_seq'), primary_key=True)
    table_name = sa.Column(sa.String(255), nullable=False)
    snapshot_id = sa.Column(
        sa.Integer, sa.ForeignKey(Snapshot.id), nullable=False
    )
    snapshot = sa.orm.relationship(Snapshot, backref='tables')

    def get_table_name(self, postfix, old=False):
        if not self.snapshot:
            raise Exception('Table name requires snapshot')
        if not self.snapshot.hash:
            raise Exception('Snapshot hash is empty.')

        if old:
            return 'stellar_%s_%s_%s' % (
                self.table_name,
                self.snapshot.hash,
                postfix
            )
        else:
            return 'stellar_%s' % hashlib.md5(
                ('%s|%s|%s' % (
                    self.table_name,
                    self.snapshot.hash,
                    postfix
                )).encode('utf-8')
            ).hexdigest()[0:16]

    def __repr__(self):
        return "<Table(table_name=%r)>" % (
            self.table_name,
        )



# coding: utf-8
import os
import re

from setuptools import setup, find_packages


# https://bitbucket.org/zzzeek/alembic/raw/f38eaad4a80d7e3d893c3044162971971ae0
# 09bf/setup.py
with open(
    os.path.join(os.path.dirname(__file__), 'stellar', 'app.py')
) as app_file:
    VERSION = re.compile(
        r".*__version__ = '(.*?)'", re.S
    ).match(app_file.read()).group(1)

with open("README.md") as readme:
    long_description = readme.read()

setup(
    name='stellar-p2p',
    description=(
        'stellar-p2p is a modified version of stellar, a tool for creating and restoring database snapshots'
    ),
    long_description=long_description,
    version=VERSION,
    url='https://github.com/craftypenguins/stellar-p2p',
    license='BSD',
    author=u'Rob Hartzenberg',
    author_email='rob@hartzenberg.net',
    packages=find_packages('.', exclude=['examples*', 'test*']),
    entry_points={
        'console_scripts': [ 'stellar-p2p = stellar.command:main' ],
    },
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Topic :: Utilities',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Database',
        'Topic :: Software Development :: Version Control',
    ],
    install_requires = [
        'PyYAML>=3.11',
        'SQLAlchemy>=0.9.6',
        'humanize>=0.5.1',
        'schema>=0.3.1',
        'click>=3.1',
        'SQLAlchemy-Utils>=0.26.11',
        'psutil>=2.1.1',
	'psycopg2>=2.7.3',
    ]
)



from stellar.models import get_unique_hash, Table, Snapshot


def test_get_unique_hash():
    assert get_unique_hash()
    assert get_unique_hash() != get_unique_hash()
    assert len(get_unique_hash()) == 32


def test_table():
    table = Table(
        table_name='hapsu',
        snapshot=Snapshot(
            snapshot_name='snapshot',
            project_name='myproject',
            hash='3330484d0a70eecab84554b5576b4553'
        )
    )
    assert len(table.get_table_name('master')) == 24
